<?php

namespace Drupal\entity_extender\Component;

use Drupal\Component\Utility\Random;
use Drupal\Component\Uuid\Php as Uuid;
use Drupal\Core\Entity\RevisionableInterface;
use Drupal\Core\Entity\EntityInterface;
use Drupal\Core\Session\AccountInterface;
use Drupal\entity_extender\Plugin\EntityExtender\ExtenderBase;

/**
 * Returns the in-RAM storage.
 *
 * @return array
 *   The in-RAM storage.
 *
 * @internal
 */
function &storage(): array {
  static $storage = [];

  return $storage;
}

/**
 * Returns the ID of an entity.
 *
 * @param mixed $input
 *   The object to get an ID of.
 *
 * @return string|null
 *   The ID of an object.
 *
 * @internal
 */
function get_entity_id($input): ?string {
  // Check for the account interface first because it's also implemented
  // by the "user" entity that is of entity interface too. This allows
  // identifying users/accounts only by email, ignoring the UUID because
  // sometimes the account is passed, sometimes - the user entity.
  if ($input instanceof AccountInterface || $input instanceof User) {
    return $input->getEmail() ?: \sprintf('%s@place.holder', (new Random())->name(8, TRUE));
  }
  elseif ($input instanceof EntityInterface || $input instanceof ExtenderBase) {
    // UUID is the same across all revisions of a specific entity, so we are
    // adding one more key - revision id, to make sure that entity id will be
    // generated per revision if applicable.
    $uuid = $input->uuid() ?: (new Uuid())->generate();
    if ($input instanceof RevisionableInterface && $revision_id = $input->getRevisionId()) {
      return \sprintf('%s_%s', $uuid, $revision_id);
    }
    return $uuid;
  }

  return NULL;
}

/**
 * The in-RAM static cache for the expensive computations.
 *
 * @link https://en.wikipedia.org/wiki/Memoization
 */
trait Memo {

  /**
   * Returns the memoized result of a callable.
   *
   * @param callable $setter
   *   The callback to produce a value for memoization.
   * @codingStandardsIgnoreStart
   * @param mixed ...$inputs
   *   The list of input parameters to associate the result with.
   * @codingStandardsIgnoreEnd
   *
   * @return mixed
   *   The memoized result.
   */
  public static function useMemo(callable $setter, ...$inputs) {
    $trace = \debug_backtrace(DEBUG_BACKTRACE_IGNORE_ARGS, 1)[0];
    // The local or trait-level static variable is not an option here because
    // for each "use" of this trait will be allocated the own storage in RAM.
    $storage =& storage();
    // Use the caller and the line of where "useMemo" is called to prevent
    // collisions.
    $pointer = [$trace['file'], $trace['line']];
    $entities = [];

    foreach ($inputs as $input) {
      if ($id = get_entity_id($input)) {
        $entities[] = $id;
      }
      // Allow passing an instance of the "ReflectionMethod" in order to
      // memoize a value for instances where the method is not overridden.
      elseif ($input instanceof \ReflectionMethod) {
        $pointer[] = $input->getDeclaringClass()->getName();
      }
      elseif (\is_scalar($input)) {
        $pointer[] = $input;
      }
      // Skip empty inputs and throw for arrays, objects, resources etc.
      elseif (!empty($input)) {
        throw new \InvalidArgumentException();
      }
    }

    // Prepend entity IDs to the pointer.
    $pointer = \array_merge($entities, $pointer);
    \assert(\count($pointer) > 2);
    $pointer = \md5(\implode('--', $pointer));

    // Store mapping of entity IDs and cache pointer.
    foreach ($entities as $id) {
      $storage['mapping'][$id][$pointer] = TRUE;
    }

    if (!\array_key_exists($pointer, $storage)) {
      $storage[$pointer] = \call_user_func($setter);
    }

    return $storage[$pointer];
  }

  /**
   * Wipes out memoized values of computations where an entity object was used.
   *
   * @param \Drupal\Core\Session\AccountInterface|\Drupal\Core\Entity\EntityInterface|\Drupal\entity_extender\Plugin\EntityExtender\ExtenderBase $input
   *   The object to invalidate the memoized values of computations where
   *   it was used.
   *
   * @see \entity_extender_entity_update()
   */
  public static function undoMemo($input): void {
    if ($id = get_entity_id($input)) {
      $storage =& storage();

      foreach ($storage['mapping'][$id] ?? [] as $pointer => $noop) {
        // Remove the data and associated cache pointer.
        unset($storage[$pointer], $storage['mapping'][$id][$pointer]);
      }
    }
  }

}
