<?php

namespace Drupal\entity_extender\Entity\Handler\Entity;

use Drupal\Core\Entity\ContentEntityInterface;
use Drupal\entity_extender\Component\Memo;
use Drupal\entity_extender\Plugin\EntityExtender\ExtenderBase;
use Drupal\entity_extender\EntityExtenderManager;

/**
 * Injects entity extender manager and allows getting specific extenders.
 *
 * @internal
 */
trait EntityHandlerEntityExtenderManagerInjector {

  use Memo;

  /**
   * An instance of the "event_dispatcher" service.
   *
   * @var \Drupal\Core\Entity\EntityFieldManagerInterface
   */
  protected $entityFieldManager;

  /**
   * An instance of the "entity_extender.manager" service.
   *
   * @var \Drupal\entity_extender\EntityExtenderManager
   */
  protected $entityExtenderManager;

  /**
   * The associative array of extenders that need extra handling.
   *
   * @var bool[]|null
   */
  protected $specificBundles;

  /**
   * A storage of the entities.
   *
   * @var \Drupal\Core\Entity\ContentEntityStorageInterface
   */
  protected $storage;

  /**
   * Returns the entity extender manager.
   *
   * @return \Drupal\entity_extender\EntityExtenderManager
   *   The entity extender manager.
   */
  protected function getEntityExtenderManager(): EntityExtenderManager {
    if ($this->entityExtenderManager === NULL) {
      $this->entityExtenderManager = \Drupal::service('entity_extender.manager');
    }

    return $this->entityExtenderManager;
  }

  /**
   * Returns a specific extender if it's preset for a bundle.
   *
   * @param string $entity_type
   *   The entity type.
   * @param string $bundle
   *   The content type.
   * @param \Drupal\Core\Entity\ContentEntityInterface $entity
   *   The entity.
   *
   * @return \Drupal\entity_extender\Plugin\EntityExtender\ExtenderBase|null
   *   The specific extender.
   */
  protected function getExtender(string $entity_type, string $bundle, ContentEntityInterface $entity): ?ExtenderBase {
    if ($interface = \constant(static::class . '::SPECIFIC_INTERFACE')) {
      $this->getEntityExtenderManager();

      if ($this->specificBundles === NULL) {
        $this->specificBundles = $this->entityExtenderManager->getImplementingBundles($entity_type, $interface);
      }

      if (isset($this->specificBundles[$bundle])) {
        $extender = static::useMemo(function () use ($entity_type, $bundle, $entity): ExtenderBase {
          return $this->entityExtenderManager->getExtender($entity_type, $bundle, $entity);
        }, $entity);

        \assert($extender instanceof $interface);

        return $extender;
      }
    }

    return NULL;
  }

}
