<?php

namespace Drupal\entity_extender;

use Drupal\Component\Plugin\Exception\PluginNotFoundException;
use Drupal\Core\Cache\CacheBackendInterface;
use Drupal\Core\Database\Connection;
use Drupal\Core\Entity\EntityInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Entity\Query\QueryBase;
use Drupal\Core\Entity\Query\QueryFactoryInterface;
use Drupal\Core\Extension\ModuleHandlerInterface;
use Drupal\Core\Session\AccountInterface;
use Drupal\node\NodeStorageInterface;
use Drupal\entity_extender\Annotation\EntityExtender;
use Drupal\entity_extender\Plugin\EntityExtender\ExtenderBase;
use Drupal\user\UserInterface;
use Drupal\user\UserStorageInterface;

/**
 * The entity extender manager.
 */
class EntityExtenderManager extends DefaultPluginManager {

  /**
   * {@inheritdoc}
   */
  protected const PLUGINS_BASE_CLASS = ExtenderBase::class;

  /**
   * {@inheritdoc}
   */
  protected const PLUGINS_ANNOTATION_CLASS = EntityExtender::class;

  /**
   * An instance of the "database" service.
   *
   * @var \Drupal\Core\Database\Connection
   */
  protected $database;

  /**
   * An instance of the "current_user" service.
   *
   * @var \Drupal\Core\Session\AccountInterface
   */
  protected $currentUser;

  /**
   * A storage of the "node" entities.
   *
   * @var \Drupal\node\NodeStorageInterface
   */
  protected $nodeStorage;

  /**
   * A storage of the "user" entities.
   *
   * @var \Drupal\user\UserStorageInterface
   */
  protected $userStorage;

  /**
   * A list of namespaces for the entity query classes.
   *
   * @var string[]
   *
   * @see \Drupal\Core\Entity\Query\Sql\QueryFactory::__construct()
   */
  protected $queryNamespaces;

  /**
   * An instance of the "entity_type.manager" service.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;

  /**
   * {@inheritdoc}
   *
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   */
  public function __construct(
    \Traversable $namespaces,
    CacheBackendInterface $cache_backend,
    ModuleHandlerInterface $module_handler,
    Connection $database,
    AccountInterface $current_user,
    QueryFactoryInterface $query_factory,
    EntityTypeManagerInterface $entity_type_manager
  ) {
    parent::__construct($namespaces, $cache_backend, $module_handler);

    $this->database = $database;
    $this->currentUser = $current_user;
    $this->queryNamespaces = QueryBase::getNamespaces($query_factory);
    $this->entityTypeManager = $entity_type_manager;
    $this->nodeStorage = $this->entityTypeManager->getStorage('node');
    $this->userStorage = $this->entityTypeManager->getStorage('user');
  }

  /**
   * Returns an instance of entity extender.
   *
   * @param string $plugin_id
   *   The ID of a plugin. Must be in "entity_type:bundle" format.
   * @param array $configuration
   *   The configuration of a plugin.
   * @param \Drupal\Core\Entity\EntityInterface $entity
   *   The entity to create an extender for.
   * @param string $class
   *   The custom class to instantiate an extender with. Must be a child
   *   of {@see \Drupal\entity_extender\Plugin\EntityExtender\ExtenderBase}.
   * @codingStandardsIgnoreStart
   * @param mixed[] ...$arguments
   *   The set of additional arguments for the extender.
   * @codingStandardsIgnoreEnd
   *
   * @return \Drupal\entity_extender\Plugin\EntityExtender\ExtenderBase
   *   An instance of the extending plugin.
   *
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   *   When the plugin definition cannot be found. {@see doGetDefinition()}.
   * @throws \InvalidArgumentException
   *   When the method tries to be used incorrectly.
   *
   * @internal
   *   Do not use this method directly as API may change! {@see getExtender()}.
   */
  public function createInstance($plugin_id, array $configuration = [], EntityInterface $entity = NULL, string $class = NULL, ...$arguments): ExtenderBase {
    if ($entity === NULL) {
      throw new \InvalidArgumentException(\sprintf('The entity for creating "%s" plugin must be passed.', $plugin_id));
    }

    if (\sprintf('%s:%s', $entity->getEntityTypeId(), $entity->bundle()) !== $plugin_id) {
      throw new \InvalidArgumentException(\sprintf('Unable to instantiate "%s" due to entity type or bundle mismatch.', $plugin_id));
    }

    $definition = $this->getDefinition($plugin_id);

    if ($class === NULL) {
      $class = $definition['class'];
    }

    if (!\is_subclass_of($class, ExtenderBase::class)) {
      throw new \InvalidArgumentException(\sprintf('The "$class" for instantiating "%s" plugin must be a child of "%s".', $plugin_id, ExtenderBase::class));
    }

    /* @var \Drupal\entity_extender\Plugin\EntityExtender\ExtenderBase $class */
    return $class::create(\Drupal::getContainer(), $definition, $configuration, $entity, $this, ...$arguments);
  }

  /**
   * {@inheritdoc}
   *
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   *   When the plugin definition cannot be found. {@see doGetDefinition()}.
   *
   * @see \Drupal\entity_extender\Annotation\EntityExtender::get()
   */
  public function getDefinition($plugin_id, $exception_on_invalid = TRUE): ?array {
    try {
      // Try standard flow and throw an exception if a plugin cannot be found.
      return parent::getDefinition($plugin_id, TRUE);
    }
    catch (PluginNotFoundException $e) {
      // Try to lookup a plugin by class.
      foreach ($this->definitions as $definition) {
        // The class is persistent since we disallow altering.
        /* @see alterDefinitions() */
        if ($definition['class'] === $plugin_id) {
          return $definition;
        }
      }

      if ($exception_on_invalid) {
        throw $e;
      }
    }

    return NULL;
  }

  /**
   * Returns entity bundles the extenders of which implements an interface.
   *
   * @param string $entity_type
   *   The type of entity.
   * @param string $interface
   *   The interface/class the extender plugin must implement/extend.
   *
   * @return bool[]
   *   The associative array where keys are bundles of the entity type and
   *   values are "TRUE".
   */
  public function getImplementingBundles(string $entity_type, string $interface): array {
    $bundles = [];

    foreach ($this->getDefinitions() as $definition) {
      /* @see \Drupal\entity_extender\Annotation\EntityExtender::get() */
      if ($definition['entity_type'] === $entity_type && \is_subclass_of($definition['class'], $interface)) {
        $bundles[$definition['entity_bundle']] = TRUE;
      }
    }

    return $bundles;
  }

  /**
   * {@inheritdoc}
   */
  public function getExtender(string $entity_type, string $bundle, EntityInterface $entity): ?ExtenderBase {
    try {
      return $this->createInstance(\sprintf('%s:%s', $entity_type, $bundle), [], $entity);
    }
    catch (\Exception $e) {
    }

    return NULL;
  }

  /**
   * Returns the extended user.
   *
   * @param null|\Drupal\Core\Session\AccountInterface $entity
   *   The entity to extend. The current user will be used if NULL.
   *
   * @return \Drupal\entity_extender\Plugin\EntityExtender\User|null
   *   The extended user or NULL if an entity type/bundle does not match.
   */
  public function getUser($entity = NULL): ?User {
    // Prevent loading already loaded entity.
    if (!$entity instanceof UserInterface) {
      // Prevent loading unsaved accounts (where ID is NULL).
      if ($entity instanceof AccountInterface && $id = $entity->id()) {
        $entity = $this->userStorage->load($id);
      }

      if (!$entity instanceof UserInterface) {
        $entity = $this->userStorage->load($this->currentUser->id());
      }
    }

    return $this->getExtender('user', 'user', $entity);
  }

  /**
   * {@inheritdoc}
   */
  public function loadUser(int $id): ?User {
    return $this->getUser($this->userStorage->load($id));
  }

  /**
   * {@inheritdoc}
   */
  public function loadUserRevision(int $revision_id): ?User {
    return $this->getUser($this->userStorage->loadRevision($revision_id));
  }

  /**
   * {@inheritdoc}
   */
  public function getNodeStorage(): NodeStorageInterface {
    return $this->nodeStorage;
  }

  /**
   * {@inheritdoc}
   */
  public function getUserStorage(): UserStorageInterface {
    return $this->userStorage;
  }

  /**
   * {@inheritdoc}
   */
  public function getEntityTypeManager(): EntityTypeManagerInterface {
    return $this->entityTypeManager;
  }

  /**
   * {@inheritdoc}
   */
  protected function alterDefinitions(&$definitions) {
    // We do not allow altering our own, custom plugins. If you
    // feel some logic needs to be changed - do it explicitly.
  }

}
