<?php

namespace Drupal\entity_extender\Plugin\EntityExtender;

use Drupal\Core\Cache\CacheableDependencyInterface;
use Drupal\Core\Entity\EntityInterface;
use Drupal\Core\Entity\FieldableEntityInterface;
use Drupal\Core\Field\EntityReferenceFieldItemListInterface;
use Drupal\Core\Logger\LoggerChannelFactoryInterface;
use Drupal\Core\Plugin\PluginBase;
use Drupal\Core\StringTranslation\TranslationInterface;
use Drupal\entity_extender\EntityExtenderManager;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * The base extender.
 */
abstract class ExtenderBase extends PluginBase implements CacheableDependencyInterface {

  /**
   * An instance of the entity.
   *
   * @var \Drupal\Core\Entity\EntityInterface
   */
  protected $entity;

  /**
   * A logger channel.
   *
   * @var \Drupal\Core\Logger\LoggerChannelInterface
   */
  protected $logger;

  /**
   * An instance of the "entity_extender.manager" service.
   *
   * @var \Drupal\entity_extender\EntityExtenderManager
   */
  protected $entityExtenderManager;

  /**
   * The internal list of values that already have been computed.
   *
   * @var array
   */
  private $values = [];

  /**
   * {@inheritdoc}
   */
  public function __construct(
    array $plugin,
    array $configuration,
    EntityInterface $entity,
    EntityExtenderManager $entity_extender_manager,
    TranslationInterface $string_translation,
    LoggerChannelFactoryInterface $logger_factory,
    // @codingStandardsIgnoreStart
    ...$arguments
    // @codingStandardsIgnoreEnd
  ) {
    \assert(isset($plugin['id']));
    \assert(\is_array($arguments));
    parent::__construct($configuration, $plugin['id'], $plugin);

    $this->entity = $entity;
    $this->logger = $logger_factory->get(\sprintf('extender:%s:%s', $this->pluginId, $this->id()));
    $this->entityExtenderManager = $entity_extender_manager;
    $this->stringTranslation = $string_translation;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(
    ContainerInterface $container,
    array $plugin,
    array $configuration,
    EntityInterface $entity,
    EntityExtenderManager $entity_extender_manager,
    // @codingStandardsIgnoreStart
    ...$arguments
    // @codingStandardsIgnoreEnd
  ) {
    return new static(
      $plugin,
      $configuration,
      $entity,
      $entity_extender_manager,
      $container->get('string_translation'),
      $container->get('logger.factory'),
      ...$arguments
    );
  }

  /**
   * {@inheritdoc}
   */
  public function __toString(): string {
    return $this->label();
  }

  /**
   * {@inheritdoc}
   */
  public function getCacheContexts(): array {
    return $this->entity->getCacheContexts();
  }

  /**
   * {@inheritdoc}
   */
  public function getCacheMaxAge(): int {
    return $this->entity->getCacheMaxAge();
  }

  /**
   * {@inheritdoc}
   */
  public function getCacheTags(): array {
    return $this->entity->getCacheTags();
  }

  /**
   * {@inheritdoc}
   */
  public function getEntity(): EntityInterface {
    return $this->entity;
  }

  /**
   * {@inheritdoc}
   */
  public function id(): ?int {
    $id = $this->entity->id();

    return $id === NULL ? $id : (int) $id;
  }

  /**
   * {@inheritdoc}
   */
  public function bundle(): string {
    return $this->entity->bundle();
  }

  /**
   * {@inheritdoc}
   */
  public function label(): string {
    return (string) $this->entity->label();
  }

  /**
   * {@inheritdoc}
   */
  public function uuid(): string {
    return (string) $this->entity->uuid();
  }

  /**
   * Returns the first referenced entity within a field.
   *
   * @param string $field_name
   *   The name of the entity reference field.
   *
   * @return \Drupal\Core\Entity\EntityInterface|static|null
   *   Either an instance of the appropriate extender or the entity object
   *   itself when no extender plugin available for this entity.
   */
  protected function getFirstReferencedEntity(string $field_name) {
    $value = $this->getReferencedEntities($field_name);

    return empty($value) ? NULL : \reset($value);
  }

  /**
   * Returns the list of referenced entities within a field.
   *
   * @param string $field_name
   *   The name of the entity reference field.
   *
   * @return \Drupal\Core\Entity\EntityInterface[]|static[]
   *   The list of referenced entities within a field. Each item is either an
   *   instance of the appropriate extender or the entity object itself when
   *   no extender plugin available for this entity.
   */
  protected function getReferencedEntities(string $field_name): array {
    return $this->computeOnce(__METHOD__ . ':' . $field_name, function () use ($field_name): array {
      if (!$this->entity instanceof FieldableEntityInterface) {
        $this->logger->warning('The "@uuid" is not a fieldable entity so cannot have references.', [
          '@uuid' => $this->uuid(),
        ]);

        return [];
      }

      $field = $this->entity->get($field_name);

      if (!$field instanceof EntityReferenceFieldItemListInterface) {
        $this->logger->warning('The "@field_name" is not a reference field of "@uuid" entity.', [
          '@field_name' => $field_name,
          '@uuid' => $this->uuid(),
        ]);

        return [];
      }

      return $this->transform($field->referencedEntities());
    });
  }

  /**
   * Returns the list of entities with applied extenders (if available).
   *
   * @param \Drupal\Core\Entity\EntityInterface[] $entities
   *   The list of entities to apply the extenders for.
   *
   * @return \Drupal\Core\Entity\EntityInterface[]|static[]
   *   The list of entities with applied extenders.
   */
  protected function transform(array $entities): array {
    $values = [];

    foreach ($entities as $entity) {
      \assert($entity instanceof EntityInterface);
      $values[$entity->uuid()] = $this->entityExtenderManager->getExtender($entity->getEntityTypeId(), $entity->bundle(), $entity) ?? $entity;
    }

    return $values;
  }

  /**
   * Computes the value once and returns it for all sequential calls.
   *
   * @param string $method
   *   The "__METHOD__" constant to ensure the value is uniquely stored. Or
   *   something you know is better.
   * @param callable $callback
   *   The callback to compute the value.
   *
   * @return mixed
   *   The computed value.
   */
  protected function computeOnce(string $method, callable $callback) {
    if (!\array_key_exists($method, $this->values)) {
      $this->values[$method] = $callback();
    }

    return $this->values[$method];
  }

  /**
   * Returns the field value of a field within an entity.
   *
   * @param string $field_name
   *   The name of a field.
   * @param string $constraint
   *   The name of a constraint to assert on a field.
   *
   * @return mixed
   *   The value of a field.
   */
  protected function get(string $field_name, string $constraint = NULL) {
    // Do no check for "hasField" here because it becomes harder to understand:
    // why this method returned "NULL" - because of missing field or missing
    // value. In case field is missing exception will be thrown.
    if ($this->entity instanceof FieldableEntityInterface) {
      // Do not use "$field->getConstraint()" since there's an "isset()" inside
      // that doesn't work when a constraint has "NULL" as a configuration.
      \assert($constraint === NULL || \array_key_exists($constraint, $this->entity->getFieldDefinition($field_name)->getConstraints()));

      return $this->entity->get($field_name)->value;
    }

    return NULL;
  }

}
