<?php

namespace Drupal\Tests\entity_extender\Unit;

use Drupal\Core\Entity\EntityInterface;
use Drupal\Core\Session\AccountInterface;
use Drupal\entity_extender\Component\Memo;
use Drupal\Tests\UnitTestCase;

/**
 * Tests memoization.
 *
 * @group entity_extender
 */
class MemoTest extends UnitTestCase {

  use Memo;

  /**
   * {@inheritdoc}
   */
  protected function getEntity(int $calls): EntityInterface {
    $entity = $this->createMock(EntityInterface::class);
    $entity
      ->expects(static::exactly($calls))
      ->method('uuid')
      ->willReturn('1f70ad30-5608-4012-b719-a71d2ad6ea0d');

    return $entity;
  }

  /**
   * {@inheritdoc}
   */
  protected function getAccount(int $calls): AccountInterface {
    $account = $this->createMock(AccountInterface::class);
    $account
      ->expects(static::exactly($calls))
      ->method('getEmail')
      ->willReturn('test@example.com');

    return $account;
  }

  /**
   * Tests memoization.
   *
   * @param bool $use_entity
   *   The state of whether to use an entity as an argument.
   * @param bool $use_account
   *   The state of whether to use an account as an argument.
   * @param \Throwable|null $exception
   *   The expected exception to be thrown during the use of a static cache.
   *
   * @dataProvider provider
   */
  public function test(bool $use_entity, bool $use_account, ?\Throwable $exception): void {
    if ($exception !== NULL) {
      $this->expectException(\get_class($exception));
      $this->expectExceptionMessage($exception->getMessage());
    }

    $iterations_count = 5;
    $executions_count = 0;

    $entity = $use_entity ? $this->getEntity($iterations_count) : NULL;
    $account = $use_account ? $this->getAccount($iterations_count) : NULL;

    for ($iterations_passed = 0; $iterations_passed < $iterations_count; $iterations_passed++) {
      static::assertSame('bla', static::useMemo(function () use (&$executions_count): string {
        $executions_count++;

        return 'bla';
      }, $entity, $account));
    }

    static::assertSame($iterations_count, $iterations_passed);
    static::assertSame(1, $executions_count);
  }

  /**
   * {@inheritdoc}
   */
  public function provider(): array {
    return [
      [FALSE, FALSE, new \AssertionError('assert(\count($pointer) > 2)')],
      [TRUE, FALSE, NULL],
      [FALSE, TRUE, NULL],
      [TRUE, TRUE, NULL],
    ];
  }

}
